#!/usr/bin/perl
# vip v2.6.3  (c) 7.5.2003 by Andreas Ley  (u) 2.8.2021
# Interface setup commands to add virtual interfaces for virtual IPs

#$use_ifconfig = 1;

# Gratuitous ARP requires installed package "iputils-arping", not "arping" !

$arp_count = 10;
$direct_routing_if = "lo";

use NetAddr::IP ':lower';

# We need to parse ifconfig output
undef $ENV{'LANG'};
undef $ENV{'LC_ALL'};
undef $ENV{'LC_MESSAGES'};

$ip_cmd = '/sbin/ip';
$use_ifconfig = 1 unless (-x $ip_cmd);
$ifconfig = '/sbin/ifconfig' if (defined($use_ifconfig));
$arping = -x '/usr/bin/arping' ? '/usr/bin/arping' : '/usr/sbin/arping';

################################################################################
#
# Check wether an IP is within the ranges of a list of networks
#
# $list: Reference to array of NetAddr::IPs
# $ip: NetAddr::IP

sub match_ip
{
	# This only shows ARRAY(0x…), but Data::Dumper would display binary address structures, interfering with terminal control characters
	&debug(__FILE__,__LINE__,'+ &match_ip('.join(',',map("'$_'",@_)).')') if (defined($debug{'call'}) && $debug{'call'}>1 || defined($debug{'match_ip'}));
	#&debug(__FILE__,__LINE__,'+ &match_ip'.Data::Dumper->new([\@_],['*'])->Indent(0)->Dump) if (defined($debug{'call'}) && $debug{'call'}>1 || defined($debug{'match_ip'}));
	my ($list,$ip) = @_;

	for my $net (@{$list}) {
		&debug(__FILE__,__LINE__,'+ &match_ip: '.$net.' -> '.($net->contains($ip))) if (defined($debug{'return'}) && $debug{'return'}>1 || defined($debug{'match_ip'}) && $debug{'match_ip'}>2);
		&debug(__FILE__,__LINE__,'+ &match_ip = '.$net) if ((defined($debug{'return'}) && $debug{'return'}>1 || defined($debug{'match_ip'}) && $debug{'match_ip'}>2) && $net->contains($ip));
		return ($net) if ($net->contains($ip));
	}

	&debug(__FILE__,__LINE__,'+ &match_ip = undef') if (defined($debug{'return'}) && $debug{'return'}>1 || defined($debug{'match_ip'}) && $debug{'match_ip'}>1);
	return (undef);
}

################################################################################
#
# Add an IP to a list of networks
#
# $list: Reference to array of NetAddr::IPs
# $ip: NetAddr::IP
# $force: Bool

sub add_ip
{
	# This only shows ARRAY(0x…), but Data::Dumper would display binary address structures, interfering with terminal control characters
	&debug(__FILE__,__LINE__,'+ &add_ip('.join(',',map("'$_'",@_)).')') if (defined($debug{'call'}) && $debug{'call'}>1 || defined($debug{'add_ip'}));
	#&debug(__FILE__,__LINE__,'+ &add_ip'.Data::Dumper->new([\@_],['*'])->Indent(0)->Dump) if (defined($debug{'call'}) && $debug{'call'}>1 || defined($debug{'add_ip'}));
	my ($list,$ip,$force) = @_;

	return if (&match_ip($list,$ip));

	&debug(__FILE__,__LINE__,"Adding $ip to $list") if (defined($debug{'add_ip'}) && $debug{'add_ip'}>1);
	push(@{$list},$ip->network);
	$force{$ip->network} = 1 if ($force);
}

################################################################################
#
# Read /etc/rc.values
#

# Direct routing ranges MUST be explicitly configured - we can't always use the fallback to $direct_routing_if if the address doesn't match any configure interface, because (e.g. in the IPv6 case) sometimes a direct routing range is a sub-subnet of a subnet configured on a routed interface!

sub init_rc_values
{
	# The old, baumschule method
	if (open(RC,"/etc/rc.values")) {
		while (<RC>) {
			chomp;
			if (/^DIRECT_ROUTING=/) {
				for (split(/ /,$')) {
					print "Using direct routing for $_\n" if (defined($opt{'verbose'}) && $opt{'verbose'}>1);
					# Use \@{ to instanciate the reference!
					&add_ip(\@{$net{$direct_routing_if}},NetAddr::IP->new($_),1);
				}
			}
			elsif (/^BIND_INTERFACE=/) {
				for (split(/ /,$')) {
					if (/:/) {
						print "Binding $' to $`\n" if (defined($opt{'verbose'}) && $opt{'verbose'}>1);
						# Use \@{ to instanciate the reference!
						&add_ip(\@{$net{$`}},NetAddr::IP->new($'),1);
					}
				}
			}
		}
		close(RC);
	}

	# The new method
	if (open(INTERFACES,"/etc/network/interfaces")) {
		my $if;
		while (<INTERFACES>) {
			chomp;
			if (/^\s*iface\s+(\S+)/) {
				$if = $1;
			}
			elsif (/^\s*bind\s+/) {
				for (split(' ',$')) {
					print "Binding $_ to $if\n" if (defined($opt{'verbose'}) && $opt{'verbose'}>1);
					# Use \@{ to instanciate the reference!
					&add_ip(\@{$net{$if}},NetAddr::IP->new($_),1);
				}
			}
		}
		close(INTERFACES);
	}
}

################################################################################
#
# Read interfaces
#

sub init_interfaces
{
	my ($if);
	if (defined($use_ifconfig)) {
		open(IFCONFIG,'-|',$ifconfig) or die "Can't run $ifconfig: $!\n";
		while (<IFCONFIG>) {
			chomp;
			if (/^(\S+)/) {
				$if = $1;
			}
			elsif (/^\s+inet addr:(\d+(?:\.\d+){3}).*Mask:(\d+(?:\.\d+){3})/ || /^\s+inet6 addr:\s+([0-9a-f]*(?::[0-9a-f]*)+)\/(\d+)/) {
				&debug(__FILE__,__LINE__,"Interface $if \"$_\"") if ($debug{'init_interfaces'});
				my $ip = NetAddr::IP->new($1,$2);
				&debug(__FILE__,__LINE__,"Interface: $if=$ip") if ($debug{'init_interfaces'});
				$if{$ip->addr} = $if;
				# As long as we use ifconfig, we have to keep track of the alias interfaces used
				$if_alias{$if} = '';
				# IPs are global to an interface and its aliases
				$if =~ s/:.*//;
				# Use \@{ to instanciate the reference!
				&add_ip(\@{$net{$if}},$ip);
			}
		}
		close(IFCONFIG);
	}
	else {
		open(IP,'-|',$ip_cmd,'addr','show') or die "Can't run $ip_cmd addr show: $!\n";
		while (<IP>) {
			chomp;
			if (/^\d+:\s*([^:]+):/) {
				$if = $1;
			}
			elsif (/^\s+inet\s+(\d+(?:\.\d+){3})\/(\d+).*\s(\S+)$/ || /^\s+inet6\s+([0-9a-f]*(?::[0-9a-f]*)+)\/(\d+)/) {
				&debug(__FILE__,__LINE__,"Interface $if \"$_\"") if ($debug{'init_interfaces'});
				my $ip = NetAddr::IP->new($1,$2);
				&debug(__FILE__,__LINE__,"Interface: $if=$ip") if ($debug{'init_interfaces'});
				$if{$ip->addr} = $if;
				# As long as we use ifconfig, we have to keep track of the alias interfaces used
				$if_alias{$3} = '' if (defined($3));
				# CHECK: Where do we need this?
				$scope{$ip->addr} = $1 if (/\sscope\s+(\S+)/);
				# Use \@{ to instanciate the reference!
				&add_ip(\@{$net{$if}},$ip);
			}
		}
		close(IP);
	}
}

################################################################################
#
# Setup virtual interfaces
#
# $gratuitous: 0=Don't send Gratuitous ARP, 1=Send Gratuitous ARP, 2=Gratuitous ARP and wait for arping to finish
# @vip: Array of NetAddr::IPs

sub add_vip
{
	&debug(__FILE__,__LINE__,'+ &add_vip('.join(',',map("'$_'",@_)).')') if (defined($debug{'call'}) || defined($debug{'add_vip'}));
	my ($gratuitous,@vip) = @_;

	for my $ip (@vip) {
		&debug(__FILE__,__LINE__,"Check $ip - defined IPs are:\n      ".join("\n      ",keys %if)) if (defined($debug{'add_vip'}) && $debug{'add_vip'}>3);
		&debug(__FILE__,__LINE__,"Check $ip - defined interfaces are:\n      ".join("\n      ",keys %net)) if (defined($debug{'add_vip'}) && $debug{'add_vip'}>3);
		unless (defined($if{$ip->addr})) {
			my @if;
			#if (defined($force)) {
			#	push(@if,$force);
			#}
			#else {
				for $if (keys %net) {
					$match = &match_ip($net{$if},$ip);
					if (defined($match)) {
						$mask = $match->mask;
						if ($force{$match}) {
							&debug(__FILE__,__LINE__,$ip->addr.': $match forced to '.$if) if (defined($debug{'add_vip'}) && $debug{'add_vip'}>2);
							@if = ($if);
							last;
						}
						push(@if,$if);
					}
				}
			#}
			#push(@if,$default) if (defined($default) && !@if);
			&debug(__FILE__,__LINE__,'Interface list for '.$ip->addr.': '.join(', ',@if)) if (defined($debug{'add_vip'}) && $debug{'add_vip'}>2);
			if (@if > 1) {
				warn "More than one interface match ",$ip->addr,": ",join(" ",@if),"\n";
			}
			else {
				#if (!@if) {
				#	warn "No suitable interface for ",$ip->addr," found!\n";
				#}
				# Instead of configuring direct routing subnets, we _default_ to direct routing
				$if = @if ? $if[0] : $direct_routing_if;
				$mask = 'host' if ($if eq $direct_routing_if);
				$ip = NetAddr::IP->new($ip->addr,$mask);
				print "Configuring ",$ip->canon," on $if\n" if (defined($opt{'verbose'}));
				my $hash;
				if ($ip->version == 4) {
					# For the sake of ifconfig users, generate an interface alias
					## We don't have to set $if_alias here since we keep track of our counter
					##do {
					##	$cnt{$if}++;
					##} while (defined($if_alias{$if.':'.$cnt{$if}}));
					# We use a hash here instead of a count since parallel execution (e.g. in init.d) leads to race conditions
					$hash = $ip->numeric % 9876;	# well, not the most sophisticated hash ;-)
					while (defined($if_alias{$if.':'.$hash})) {
						$hash++;
					}
					$if_alias{$if.':'.$hash} = '';
				}
				if (defined($use_ifconfig)) {
					if ($ip->version == 4) {
						##@config = ($ifconfig,$if.':'.$cnt{$if},$ip->addr,'broadcast',$ip->broadcast->addr,'netmask',$ip->mask);
						@config = ($ifconfig,$if.':'.$hash,$ip->addr,'broadcast',$ip->broadcast->addr,'netmask',$ip->mask);
					}
					else {
						@config = ($ifconfig,$if,'add',$ip);
					}
				}
				else {
					@config = ($ip_cmd,'addr','add',$ip,'dev',$if);	# brd $ip->broadcast->addr
					push(@config,'label',$if.':'.$hash) if ($ip->version == 4);
				}
				$tentative = 1 if ($ip->version == 6);
				my @arping = ($arping,'-U','-q','-c',$arp_count,'-I',$if,$ip->addr) if ($gratuitous);
				if (defined($opt{'dry-run'})) {
					print join(' ',@config),"\n";
					print join(' ',@arping),"\n" if (@arping);
				}
				else {
					# FIXME: Redirect to STDERR
					system(defined($use_ifconfig)?$ifconfig:($ip_cmd,'addr','show')) if (defined($debug{'add_vip'}) && $debug{'add_vip'}>3);
					&debug(__FILE__,__LINE__,'+ '.join(' ',@config)) if (defined($debug{'exec'}) || defined($debug{'add_vip'}) && $debug{'add_vip'}>2);
					system(@config) and die 'Failed to run '.join(' ',@config)."\n";
					system(defined($use_ifconfig)?$ifconfig:($ip_cmd,'addr','show')) if (defined($debug{'add_vip'}) && $debug{'add_vip'}>3);
					if ($gratuitous) {
						# Synchroneous arping waits sequentially :-(
						#system(@arping) and die "Failed to run ".join(' ',@arping)."\n";
						die "Can't fork: $!\n" unless (defined($pid = fork));
						if ($pid) {
							print "Child process $pid started for ".join(' ',@arping)."\n" if ($debug{'add_vip'});
							if ($gratuitous>1) {
								push(@wait,$pid);
								$cmd{$pid} = join(' ',@arping);
							}
						}
						else {
							print "Child process running ".join(' ',@arping)."\n" if ($debug{'add_vip'});
							&debug(__FILE__,__LINE__,'+ '.join(' ',@arping)) if (defined($debug{'exec'}) || defined($debug{'add_vip'}) && $debug{'add_vip'}>2);
							exec(@arping);
							die 'Failed to exec '.join(' ',@arping)."\n";
						}
					}
				}
				# Store what should have happend, in case there are multiple calls for the same IP
				$if{$ip->addr} = $if;
			}
		}
		elsif (defined($opt{'verbose'})) {
			print $ip->canon," already defined on ",$if{$ip->addr},"\n";
		}
	}

	if ($tentative) {
		if (defined($use_ifconfig)) {
			warn "IPv6 addresses may still be tentative!\nUnfortunatly, this cannot be seen with ifconfig :-(\n";
		}
		else {
			system("while $ip_cmd addr show tentative | grep -q tentative; do sleep 1; done");
		}
	}
	if ($gratuitous>1) {
		for my $pid (@wait) {
			warn "Failed to run ".$cmd{$pid}.": return code $?\n" if (waitpid($pid,0) && $?);
		}
	}
}

################################################################################
#
# Unconfigure virtual interfaces
#
# @vip: Array of NetAddr::IPs

sub del_vip
{
	&debug(__FILE__,__LINE__,'+ &del_vip('.join(',',map("'$_'",@_)).')') if (defined($debug{'call'}) || defined($debug{'del_vip'}));
	my (@vip) = @_;

	for my $ip (@vip) {
		# CHECK: If we want matching regardless of mask, we should use $if{$ip->addr} everywhere (or $ip->canon ?)
		if (defined($if{$ip->addr})) {
				print "Unconfiguring ",$ip->canon," from ",$if{$ip->addr},"\n" if (defined($opt{'verbose'}));
				my @config;
				if (defined($use_ifconfig)) {
					@config = ($ifconfig,$if{$ip->addr},(NetAddr::IP->new($ip->addr)->version==4?'down':('del',$ip->addr)));
				}
				else {
					@config = ($ip_cmd,'addr','del',$ip->cidr,'dev',$if{$ip->addr});
				}
				if (defined($opt{'dry-run'})) {
					print join(' ',@config)."\n";
				}
				else {
					&debug(__FILE__,__LINE__,'+ '.join(' ',@config)) if (defined($debug{'exec'}) || defined($debug{'del_vip'}) && $debug{'del_vip'}>2);
					system(@config) and die 'Failed to run '.join(' ',@config)."\n";
				}
		}
		elsif (defined($opt{'verbose'})) {
			print $ip->canon," not configured on any interface\n";
		}
	}
}

1;
